// pocketpcdll.cpp : Defines the initialization routines for the DLL.
//
#include "stdafx.h"
#include "SpeexOutLoud.h"
#include "speex.h"
#include <winbase.h>
#include <stdio.h>
#include <string.h>
#include <STDLIB.H>
#include <windows.h>
#include "nb_celp.h"
#include "misc.h"
#include "modes.h"
#include <malloc.h>

#ifdef _DEBUG
#define new DEBUG_NEW
#undef THIS_FILE
static char THIS_FILE[] = __FILE__;
#endif

#define FRAME_SIZE 160
#define WAVEHEADER 60

#define __attribute__(x)
#define sleep(__sec)        Sleep ((__sec) * 1000)
#define inline              __inline
#define restrict


#define MAXWAVESIZE     4294967040LU
#define MAX_WAVEBLOCKS    32

#define CD_SAMPLE_FREQ         44.1e3
#define SAMPLE_SIZE            16
#define SAMPLE_SIZE_STRING     ""
#define WINAUDIO_FD            ((FILE_T)-128)
#define FILE_T                 FILE*
#define INVALID_FILEDESC       NULL

typedef signed   int        Int;        // at least -32767...+32767, fast type
typedef unsigned int        Uint;       // at least 0...65535, fast type
typedef long double         Ldouble;    // most exact floating point format

static CRITICAL_SECTION  cs;
static HWAVEOUT          dev                    = NULL;
static int               ScheduledBlocks        = 0;
static int               PlayedWaveHeadersCount = 0;          // free index
static WAVEHDR*          PlayedWaveHeaders [MAX_WAVEBLOCKS];

int        WIN_Play_Samples           ( const void* buff, size_t len );
int        WIN_Audio_close            ( void );
Int        SetParams             ( FILE_T dummyFile , Ldouble SampleFreq, Uint BitsPerSample, Uint Channels);
//
//	Note!
//
//		If this DLL is dynamically linked against the MFC
//		DLLs, any functions exported from this DLL which
//		call into MFC must have the AFX_MANAGE_STATE macro
//		added at the very beginning of the function.
//
//		For example:
//
//		extern "C" BOOL PASCAL EXPORT ExportedFunction()
//		{
//			AFX_MANAGE_STATE(AfxGetStaticModuleState());
//			// normal function body here
//		}
//
//		It is very important that this macro appear in each
//		function, prior to any calls into MFC.  This means that
//		it must appear as the first statement within the 
//		function, even before any object variable declarations
//		as their constructors may generate calls into the MFC
//		DLL.
//
//		Please see MFC Technical Notes 33 and 58 for additional
//		details.
//

/////////////////////////////////////////////////////////////////////////////
// CSpeexOutLoudApp

BEGIN_MESSAGE_MAP(CSpeexOutLoudApp, CWinApp)
	//{{AFX_MSG_MAP(CSpeexOutLoudApp)
		// NOTE - the ClassWizard will add and remove mapping macros here.
		//    DO NOT EDIT what you see in these blocks of generated code!
	//}}AFX_MSG_MAP
END_MESSAGE_MAP()

/////////////////////////////////////////////////////////////////////////////
// CSpeexOutLoudApp construction

CSpeexOutLoudApp::CSpeexOutLoudApp()
{
	// TODO: add construction code here,
	// Place all significant initialization in InitInstance
}

/////////////////////////////////////////////////////////////////////////////
// The one and only CSpeexOutLoudApp object

CSpeexOutLoudApp theApp;

static void free_memory ( void )
{
	WAVEHDR*  wh;
	HGLOBAL   hg;

	EnterCriticalSection ( &cs );
	wh = PlayedWaveHeaders [--PlayedWaveHeadersCount];
	ScheduledBlocks--;                               // decrease the number of USED blocks
	LeaveCriticalSection ( &cs );

	waveOutUnprepareHeader ( dev, wh, sizeof (WAVEHDR) );

	hg = GlobalHandle ( wh -> lpData );       // Deallocate the buffer memory
	GlobalUnlock (hg);
	GlobalFree   (hg);

	hg = GlobalHandle ( wh );                 // Deallocate the header memory
	GlobalUnlock (hg);
	GlobalFree   (hg);
}

/* Input param - FileName -  Path of the Speex file to be played on the 
 Pocket PC */

_declspec(dllexport) int PlaySpeexFiles(LPTSTR FileName)
{
	
SpeexBits bits; 
 
 void *state; 

 char InBuffer[200]; 
 short OutBuffer[FRAME_SIZE]; 
 char waveheader[WAVEHEADER]; 
 
 HANDLE InFile; 
 HANDLE OutFile; 
 
 int RetResult;
 char nrBytes = 0; 
 int nbBytes = 0; 
 
DWORD NrOfBytesRead; 
DWORD NrOfBytesWritten; 
 
 
 state = speex_decoder_init(&speex_nb_mode); //Init decoder in //narrowband mode 
 
 InFile = CreateFile(FileName, GENERIC_READ, 0, NULL, OPEN_EXISTING, FILE_ATTRIBUTE_NORMAL, NULL);
  
 SetParams (INVALID_FILEDESC, 8000, 16, 1);
 
 int tmp=1;  //Set perpetual enhancement on
 
 speex_decoder_ctl(state, SPEEX_SET_ENH, &tmp); 
 
 speex_bits_init(&bits); //Init out bit buffer 
 
 ReadFile(InFile, waveheader, WAVEHEADER, &NrOfBytesRead, NULL); //Raw copy of wave header 
 
 ReadFile(InFile, &nrBytes, 1, &NrOfBytesRead, NULL); //Read first byte in frame to "nrBytes" 
 
 nbBytes=nrBytes; //"nrBytes" gives info about used code rate 
 
  while(1){ //Decode until end of file
      ReadFile(InFile, InBuffer, nbBytes, &NrOfBytesRead, NULL); //Read bits to decode 
  
      speex_bits_read_from(&bits, InBuffer, nbBytes); 
 
      speex_decode_int(state, &bits, OutBuffer); //Decode bits 
 
	  RetResult=WIN_Play_Samples (OutBuffer, sizeof(short) * FRAME_SIZE);

      if(NrOfBytesRead<nbBytes) //Break if end of file 
           break;
 } 


 //Deallocate 
 speex_bits_destroy(&bits); 
 speex_decoder_destroy(state); 
 CloseHandle(InFile);
 WIN_Audio_close ();
 return 0;
}

static void CALLBACK wave_callback ( HWAVE hWave, UINT uMsg, DWORD dwInstance, DWORD dwParam1, DWORD dwParam2 )
{
	if ( uMsg == WOM_DONE ) {
		EnterCriticalSection ( &cs );
		PlayedWaveHeaders [PlayedWaveHeadersCount++] = (WAVEHDR*) dwParam1;
		LeaveCriticalSection ( &cs );
	}
}

Int
SetParams ( FILE_T   dummyFile ,
                 Ldouble  SampleFreq,
                 Uint     BitsPerSample,
                 Uint     Channels )
{
	WAVEFORMATEX  outFormat;
	UINT          deviceID = WAVE_MAPPER;

	(void) dummyFile;

	if ( waveOutGetNumDevs () == 0 )
		return -1;

	outFormat.wFormatTag      = WAVE_FORMAT_PCM;
	outFormat.wBitsPerSample  = BitsPerSample;
	outFormat.nChannels       = Channels;
	outFormat.nSamplesPerSec  = (unsigned long)(SampleFreq + 0.5);
	outFormat.nBlockAlign     = (outFormat.wBitsPerSample + 7) / 8 * outFormat.nChannels;
	outFormat.nAvgBytesPerSec = outFormat.nSamplesPerSec * outFormat.nBlockAlign;

	switch ( waveOutOpen ( &dev, deviceID, &outFormat, (DWORD)wave_callback, 0, CALLBACK_FUNCTION ) )
	{
		case MMSYSERR_ALLOCATED:   return -1;
		case MMSYSERR_BADDEVICEID: return -1;
		case MMSYSERR_NODRIVER:    return -1;
		case MMSYSERR_NOMEM:       return -1;
		case WAVERR_BADFORMAT:     return -1;
		case WAVERR_SYNC:          return -1;
		default:                   return -1;
		case MMSYSERR_NOERROR:     break;
	}

	waveOutReset ( dev );
	InitializeCriticalSection ( &cs );	
	return 0;
}

int
WIN_Play_Samples (const void* data, size_t len )
{
	HGLOBAL    hg;
	HGLOBAL    hg2;
	LPWAVEHDR  wh;
	void*      allocptr;

	do {
		while ( PlayedWaveHeadersCount > 0 )                // free used blocks ...
			free_memory ();

		if ( ScheduledBlocks < sizeof(PlayedWaveHeaders)/sizeof(*PlayedWaveHeaders) ) // wait for a free block ...
			break;
		Sleep (26);
	} while (1);

	if ( (hg2 = GlobalAlloc ( GMEM_MOVEABLE, len )) == NULL )   // allocate some memory for a copy of the buffer
		return -1;

	allocptr = GlobalLock (hg2);

	CopyMemory ( allocptr, data, len );                        // Here we can call any modification output functions we want....
	

	if ( (hg = GlobalAlloc (GMEM_MOVEABLE | GMEM_ZEROINIT, sizeof (WAVEHDR))) == NULL ) // now make a header and WRITE IT!
		return -1;

	wh                   = (LPWAVEHDR)GlobalLock (hg);
	wh -> dwBufferLength = len;
	wh -> lpData         = (char*)allocptr;

	if ( waveOutPrepareHeader ( dev, wh, sizeof (WAVEHDR)) != MMSYSERR_NOERROR ) {
		GlobalUnlock (hg);
		GlobalFree   (hg);
		return -1;
	}

	if ( waveOutWrite ( dev, wh, sizeof (WAVEHDR)) != MMSYSERR_NOERROR ) {
		GlobalUnlock (hg);
		GlobalFree   (hg);
		return -1;
	}

	EnterCriticalSection ( &cs );
	ScheduledBlocks++;
	LeaveCriticalSection ( &cs );

	return len;
}

int
WIN_Audio_close ( void )
{
	if ( dev != NULL ) {

		while ( ScheduledBlocks > 0 ) {
			Sleep (ScheduledBlocks);
			while ( PlayedWaveHeadersCount > 0 )         // free used blocks ...
				free_memory ();
		}

		waveOutReset (dev);      // reset the device
		waveOutClose (dev);      // close the device
		dev = NULL;
	}

	DeleteCriticalSection ( &cs );
	ScheduledBlocks = 0;
	return 0;
}

/* Input param(1) - FileName - Name of the WAV file to be encoded on the
   Pocket PC.
   Input param (2) - EncFileName - Name of the encoded .spx file on the Pocket PC
*/
_declspec(dllexport) int EncodeWaveFiles(LPTSTR FileName,LPTSTR EncFileName)
{
	SpeexBits bits;
	void *state;
	short InBuffer[FRAME_SIZE];
	short InFloat[FRAME_SIZE];
	char OutBuffer[200];
	char waveheader[WAVEHEADER];
	int tmp;
	HANDLE WaveFile;
	HANDLE SpeexFile;

	DWORD NrofBytesRead,NrOfBytesWritten;

	char nrBytes=0;
	int nbBytes=0;
	int i,bit_rate,frame_size,test=0;
	
	bool first=true;
	
	WaveFile=CreateFile(FileName,GENERIC_READ,0,NULL,OPEN_EXISTING,FILE_ATTRIBUTE_NORMAL,NULL);

	SpeexFile=CreateFile(EncFileName,GENERIC_WRITE,0,NULL,CREATE_ALWAYS,FILE_ATTRIBUTE_NORMAL,NULL);

	state=speex_encoder_init(&speex_nb_mode);
	
	tmp=10; //set quality on
	
	speex_encoder_ctl(state,SPEEX_SET_QUALITY,&tmp);
	
	speex_encoder_ctl(state,SPEEX_GET_FRAME_SIZE,&frame_size);

	speex_bits_init(&bits);

	ReadFile(WaveFile,waveheader,WAVEHEADER,&NrofBytesRead,NULL);

	WriteFile(SpeexFile,waveheader,WAVEHEADER,&NrOfBytesWritten,NULL);

	while(1)
	{
		ReadFile(WaveFile,InBuffer,FRAME_SIZE*2, &NrofBytesRead,NULL);

		for(i=0;i<FRAME_SIZE;i++)
			InFloat[i]=InBuffer[i];

		speex_bits_reset(&bits);

		speex_encode_int(state,InFloat,&bits);

		nbBytes=speex_bits_write(&bits,OutBuffer,200);

		if(first)
		{
			nrBytes=nbBytes;
			WriteFile(SpeexFile,&nrBytes,1,&NrOfBytesWritten,NULL);
			first=false;
		}
		WriteFile(SpeexFile,&OutBuffer,nbBytes,&NrOfBytesWritten,NULL);

		if(NrofBytesRead<FRAME_SIZE)
			break;
	}

	speex_bits_destroy(&bits);
	speex_encoder_destroy(state);
	
	CloseHandle(WaveFile);
	CloseHandle(SpeexFile);
	return 0;
}
